package org.artifactory.engine;

import org.apache.http.HttpHeaders;
import org.artifactory.addon.AddonsManager;
import org.artifactory.aol.AolService;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.config.CentralConfigDescriptorImpl;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.fs.RepoResource;
import org.artifactory.repo.Repo;
import org.artifactory.repo.RepoPath;
import org.artifactory.request.ArtifactoryRequest;
import org.artifactory.request.InternalRequestContext;
import org.artifactory.request.RequestClientType;
import org.artifactory.resource.RepoResourceInfo;
import org.artifactory.security.access.AccessTokenAuthentication;
import org.artifactory.storage.binstore.service.InternalBinaryService;
import org.jfrog.access.token.JwtAccessToken;
import org.jfrog.storage.binstore.common.BinaryElementRequestImpl;
import org.jfrog.storage.binstore.ifc.model.BinaryElementRequest;
import org.jfrog.storage.binstore.providers.base.ManualRedirectType;
import org.jfrog.storage.binstore.providers.base.RedirectType;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.util.Map;

import static org.jfrog.storage.binstore.ifc.model.BinaryElementHeaders.SKIP_CONTENT_DISPOSITION;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.clearInvocations;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.AssertJUnit.assertEquals;

/**
 * @author Shay Bagants
 */
@Test
public class DownloadServiceImplTest {

    private final DownloadServiceImpl downloadService = spy(new DownloadServiceImpl());

    @Mock
    private InternalRequestContext requestContext;
    @Mock
    private ArtifactoryRequest request;
    @Mock
    private AuthorizationService authorizationService;
    @Mock
    private DownloadRedirectHelper downloadRedirectHelper;
    @Mock
    private InternalBinaryService binaryService;
    @Mock
    private CentralConfigService centralConfig;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private AolService aolService;

    @DataProvider
    public static Object[][] provideRequestData() {
        return new Object[][] {
                //agent, user, expectedResult
                {"xray/1.0", "xray", "xray"},
                {"xray/1.0", "test-user", null},
                {"jfrog-cli", "xray", null},
                {"xray/1.0", "token:jfrt@123", null},
                {"jfrog-cli", "token:jfxr@123", null},
                {"x-ray/1.0", "token:jfxr@123", null},
        };
    }

    @DataProvider
    public static Object[][] provideRequestDataForAccessTokenAuthentication() {
        return new Object[][] {
                //agent, user, expectedResult
                {"xray/1.0", "xray", null},
                {"xray/1.0", "test-user", null},
                {"jfrog-cli", "xray", null},
                {"xray/1.0", "token:jfrt@123", null},
                {"jfrog-cli", "jfxr@123", null},
                {"xray/1.0", "jfxr@123", "jfxr@123"},
                {"x-ray/1.0", "jfxidx@123", null},
                {"xray/1.0", "jfxidx@123", "jfxidx@123"},
                {"x-ray/1.0", "jfxpst@123", null},
                {"xray/1.0", "jfxpst@123", "jfxpst@123"},
                {"x-ray/1.0", "jfxana@123", null},
                {"xray/1.0", "jfxana@123", "jfxana@123"},
        };
    }

    @DataProvider
    public static Object[][] getCdnConfigurationsData() {
        return new Object[][] {
                //isCdn, isEplus, isEdge, isEnterprise, isAol, isAnonymous, manualSentType, expectedType
                //AOL
                {true, true, false, false, true, false, ManualRedirectType.CF, RedirectType.CLOUD_FRONT},
                {true, true, false, false, true, false, ManualRedirectType.DEFAULT, RedirectType.CLOUD_FRONT},
                {true, false, true, false, true, false, ManualRedirectType.CF, RedirectType.CLOUD_FRONT},
                {true, true, false, false, true, true, ManualRedirectType.CF, RedirectType.CLOUD_FRONT},
                {true, false, true, false, true, true, ManualRedirectType.CF, RedirectType.CLOUD_FRONT},
                {true, false, false, true, true, true, ManualRedirectType.CF, RedirectType.CLOUD_FRONT},
                {true, false, false, true, true, false, ManualRedirectType.S3, RedirectType.S3},
                {true, false, false, true, true, false, ManualRedirectType.DEFAULT, RedirectType.S3},
                {true, false, false, false, true, true, ManualRedirectType.S3, RedirectType.S3},
                {true, false, false, false, true, false, ManualRedirectType.S3, RedirectType.S3},
                {false, true, false, false, true, false, ManualRedirectType.S3, RedirectType.S3},
                {false, false, true, false, true, false, ManualRedirectType.S3, RedirectType.S3},
                {false, false, false, false, true, false, ManualRedirectType.S3, RedirectType.S3},
                //OnPrem - isCdn cannot be true onPrem (as of 29/03/20)
                {false, true, false, false, false, false, ManualRedirectType.S3, RedirectType.S3},
                {false, false, true, false, false, false, ManualRedirectType.S3, RedirectType.S3},
                {false, false, false, false, false, false, ManualRedirectType.NONE, RedirectType.DIRECT},
                {false, false, false, true, false, true, ManualRedirectType.NONE, RedirectType.DIRECT},
                {false, false, false, true, false, true, ManualRedirectType.DEFAULT, RedirectType.DIRECT},
                {false, false, false, true, false, false, ManualRedirectType.NONE, RedirectType.DIRECT},
                //Manual DIRECT request - will return direct when it's not AOL
                {false, true, false, false, false, false, ManualRedirectType.NONE, RedirectType.DIRECT},
                {false, true, false, false, false, false, ManualRedirectType.NONE, RedirectType.DIRECT},
                //Manual DIRECT request - will ignore the manual request when it's AOL
                {true, true, false, false, true, false, ManualRedirectType.NONE, RedirectType.CLOUD_FRONT},
                {false, false, true, false, true, false, ManualRedirectType.NONE, RedirectType.S3},
        };
    }

    @BeforeMethod
    private void setup() {
        MockitoAnnotations.openMocks(this);
        when(requestContext.getRequest()).thenReturn(request);

        downloadService.authorizationService = authorizationService;
        downloadService.downloadRedirectHelper = downloadRedirectHelper;
        downloadService.addonsManager = addonsManager;
        downloadService.binaryService = binaryService;
        downloadService.centralConfig = centralConfig;

        ArtifactoryContext context = mock(ArtifactoryContext.class);
        when(context.beanForType(AolService.class)).thenReturn(aolService);
        ArtifactoryContextThreadBinder.bind(context);
    }

    @AfterMethod
    private void unbind() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test(dataProvider = "provideRequestData")
    public void testGetXrayUserName(String userAgent, String principal, String expectedResult) {
        when(authorizationService.currentUsername()).thenReturn(principal);
        when(request.getHeader(HttpHeaders.USER_AGENT)).thenReturn(userAgent);
        String result = downloadService.getXrayUserName(requestContext, null);
        Assert.assertEquals(result, expectedResult);
    }

    @Test(dataProvider = "provideRequestDataForAccessTokenAuthentication", description = "RTFACT-20177")
    public void testGetXrayUserNameUsingAccessTokenAuthentication(String userAgent, String principal,
            String expectedResult) {
        AccessTokenAuthentication accessTokenAuthentication = mock(AccessTokenAuthentication.class);
        JwtAccessToken jwtAccessToken = mock(JwtAccessToken.class);
        when(accessTokenAuthentication.getAccessToken()).thenReturn(jwtAccessToken);
        when(jwtAccessToken.getSubject()).thenReturn(principal);
        when(request.getHeader(HttpHeaders.USER_AGENT)).thenReturn(userAgent);
        String result = downloadService.getXrayUserName(requestContext, accessTokenAuthentication);
        Assert.assertEquals(result, expectedResult);
    }

    @Test(dataProvider = "getCdnConfigurationsData")
    public void testGetRedirectTypeForRepo_differentCdnConfigurations(boolean isCdn, boolean isEplus, boolean isEdge,
            boolean isEnterprise, boolean isAol, boolean isAnonymous, ManualRedirectType typeSent,
            RedirectType expectedType) {
        setupCdn();

        RepoResource resource = mock(RepoResource.class);
        Repo repo = mock(Repo.class);
        LocalRepoDescriptor repoDescriptor = mock(LocalRepoDescriptor.class);
        RepoResourceInfo repoResourceInfo = mock(RepoResourceInfo.class);

        when(repoResourceInfo.getSha1()).thenReturn("cf23df2207d99a74fbe169e3eba035e633b65d94");
        when(resource.getInfo()).thenReturn(repoResourceInfo);
        when(repoDescriptor.isReal()).thenReturn(true);
        when(repo.getDescriptor()).thenReturn(repoDescriptor);

        // using params
        when(repoDescriptor.isCdnRedirect()).thenReturn(isCdn);
        when(addonsManager.isEnterprisePlusInstalled()).thenReturn(isEplus);
        when(addonsManager.isBasicEnterpriseLicensed()).thenReturn(isEnterprise);
        when(addonsManager.isEdgeLicensed()).thenReturn(isEdge);
        when(aolService.isAol()).thenReturn(isAol);
        when(authorizationService.isAnonymous()).thenReturn(isAnonymous);
        RedirectType redirectType = downloadService.getRedirectTypeForRepo(requestContext, repo, resource, typeSent);

        assertEquals(expectedType, redirectType);
    }

    @Test
    public void testGetSignUrl() {
        // setup
        RepoResource resource = mock(RepoResource.class);
        RepoResourceInfo repoResourceInfo = mock(RepoResourceInfo.class);
        when(repoResourceInfo.getSha1()).thenReturn("cf23df2207d99a74fbe169e3eba035e633b65d94");
        when(resource.getInfo()).thenReturn(repoResourceInfo);
        Map<String, String> headers = downloadService
                .getHeaderMapForRedirect(repoResourceInfo, "user", RedirectType.S3, RequestClientType.DEBIAN,
                        ManualRedirectType.DEFAULT);
        BinaryElementRequest request = new BinaryElementRequestImpl(resource.getInfo().getSha1(), headers);
        when(binaryService.getSignedUrl(request)).thenReturn("http://signed.url");

        // test
        String url = downloadService
                .getSignedUrl(resource, "user", RedirectType.S3, RequestClientType.DEBIAN, ManualRedirectType.DEFAULT);

        // verify
        assertEquals("http://signed.url", url);
    }

    @Test
    public void testGetHeaderMapForRedirect() {
        // setup
        RepoPath repoPath = mock(RepoPath.class);
        when(repoPath.getName()).thenReturn("repoName");
        when(repoPath.getPath()).thenReturn("repoPath");

        RepoResourceInfo resourceInfo = mock(RepoResourceInfo.class);
        when(resourceInfo.getRepoPath()).thenReturn(repoPath);

        // test
        Map<String, String> headers = downloadService
                .getHeaderMapForRedirect(resourceInfo, "user", RedirectType.S3, RequestClientType.UNKNOWN,
                        ManualRedirectType.DEFAULT);

        // verify
        Assert.assertEquals(headers.get(SKIP_CONTENT_DISPOSITION.getHeaderName()), "false");

        // test
        headers = downloadService
                .getHeaderMapForRedirect(resourceInfo, "user", RedirectType.S3, RequestClientType.DEBIAN,
                        ManualRedirectType.DEFAULT);

        // verify
        Assert.assertEquals(headers.get(SKIP_CONTENT_DISPOSITION.getHeaderName()), "true");

        // test
        headers = downloadService
                .getHeaderMapForRedirect(resourceInfo, "user", RedirectType.S3, RequestClientType.APACHE_IVY,
                        ManualRedirectType.DEFAULT);

        // verify
        Assert.assertEquals(headers.get(SKIP_CONTENT_DISPOSITION.getHeaderName()), "false");
    }

    @Test(dataProvider = "shouldRedirectProvider")
    public void testShouldRedirect(boolean headRequest, boolean expectedResult) {
        LocalRepoDescriptor repoDescriptor = mock(LocalRepoDescriptor.class);
        when(repoDescriptor.isReal()).thenReturn(true);
        when(downloadRedirectHelper.isRedirectEnable(any(), any())).thenReturn(true);
        when(downloadRedirectHelper.isArtifactGreaterEqualsRedirectThreshold(any())).thenReturn(true);
        when(binaryService.canCloudProviderGenerateRedirection(any())).thenReturn(true);
        when(request.isHeadOnly()).thenReturn(headRequest);
        RepoResource resource = mock(RepoResource.class);
        RepoResourceInfo repoResourceInfo = mock(RepoResourceInfo.class);
        when(repoResourceInfo.getSha1()).thenReturn("cf23df2207d99a74fbe169e3eba035e633b65d94");
        when(resource.getInfo()).thenReturn(repoResourceInfo);

        boolean shouldRedirect = downloadService.shouldRedirect(requestContext, null, resource, repoDescriptor, true);
        assertEquals(shouldRedirect, expectedResult);
    }

    @Test
    public void testOnContextLoadedNotifyRepoRedirectAllowance() {
        // setup
        CentralConfigDescriptorImpl descriptor = new CentralConfigDescriptorImpl();
        Map<String, LocalRepoDescriptor> localRepoMap = descriptor.getLocalRepositoriesMap();
        Map<String, FederatedRepoDescriptor> federatedRepoMap = descriptor.getFederatedRepositoriesMap();
        String repoKey1 = "gen1";
        String repoKey2 = "gen2";
        LocalRepoDescriptor gen1 = createFederatedOrLocalRepo(repoKey1, false);
        FederatedRepoDescriptor gen2 = (FederatedRepoDescriptor) createFederatedOrLocalRepo(repoKey2, true);
        localRepoMap.put(repoKey1, gen1);
        federatedRepoMap.put(repoKey2, gen2);
        when(centralConfig.getDescriptor()).thenReturn(descriptor);

        // test
        downloadService.onLicenseLoaded();
        // verify non have redirect enabled by default
        verify(downloadService, times(0)).logWarnRedirectConfigured();

        // setup
        gen1.setDownloadRedirect(true);
        // test
        downloadService.onLicenseLoaded();
        // verify non have redirect enabled by default
        verify(downloadService, times(1)).logWarnRedirectConfigured();
        clearInvocations(downloadService);

        // setup
        gen2.setDownloadRedirect(true);
        // test
        downloadService.onLicenseLoaded();
        // verify non have redirect enabled by default
        verify(downloadService, times(1)).logWarnRedirectConfigured();
        clearInvocations(downloadService);
    }

    private LocalRepoDescriptor createFederatedOrLocalRepo(String repoKey, boolean federated) {
        LocalRepoDescriptor repoDescriptor;
        if (federated) {
            repoDescriptor = new FederatedRepoDescriptor();
        } else {
            repoDescriptor = new LocalRepoDescriptor();
        }
        repoDescriptor.setKey(repoKey);
        repoDescriptor.setType(RepoType.Generic);
        return repoDescriptor;
    }

    @DataProvider
    public Object[][] shouldRedirectProvider() {
        return new Object[][] {
                {true, false},
                {false, true}
        };
    }

    private void setupCdn() {
        when(downloadRedirectHelper.isRedirectEnable(any(), any())).thenReturn(true);
        when(downloadRedirectHelper.isArtifactGreaterEqualsRedirectThreshold(any())).thenReturn(true);
        when(binaryService.canCloudProviderGenerateRedirection(any())).thenReturn(true);
        when(aolService.isAol()).thenReturn(true);
    }
}