package org.artifactory.ui.rest.service.admin.configuration.repositories.util.builder;

import com.google.common.collect.Maps;
import org.artifactory.aol.AolService;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.property.ArtifactorySystemProperties;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.ProxyDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.VirtualRepoDescriptor;
import org.artifactory.engine.DownloadRedirectHelper;
import org.artifactory.storage.binstore.service.BinaryService;
import org.artifactory.ui.rest.model.admin.configuration.repository.BasicRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.CdnRedirectRepoConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.GeneralRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.federation.FederatedRepoAdvancedRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.federation.FederatedRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.federation.FederationConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.local.LocalAdvancedRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.local.LocalBasicRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.local.LocalRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteAdvancedRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteBasicRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteCacheRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.DockerTypeSpecificConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.NpmTypeSpecificConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.TypeSpecificConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.virtual.VirtualBasicRepositoryConfigModel;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.Map;

import static org.artifactory.repo.config.RepoConfigDefaultValues.DEFAULT_DOCKER_REMOTE_RETRIEVAL_CACHE_PERIOD;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertTrue;

/**
 * Author: Ivan Perepelytsya
 * Date: 6/3/20.
 */
@Test
public class RepoConfigDescriptorBuilderTest {

    @Mock
    private DownloadRedirectHelper downloadRedirectHelper;
    @Mock
    private BinaryService binaryService;
    @Mock
    private ArtifactoryContext artifactoryContext;
    @Mock
    private ArtifactoryHome artifactoryHome;
    @Mock
    private CentralConfigService centralConfigService;
    @Mock
    private MutableCentralConfigDescriptor mutableCentralConfigDescriptor;
    @Mock
    private ArtifactorySystemProperties artifactorySystemProperties;
    @InjectMocks
    private RepoConfigDescriptorBuilder repoConfigDescriptorBuilder;
    private TypeSpecificConfigModel typeSpecificConfigModel;

    @BeforeMethod
    public void initMocks() {
        MockitoAnnotations.initMocks(this);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);
        ArtifactoryHome.bind(artifactoryHome);
        when(artifactoryContext.getCentralConfig()).thenReturn(centralConfigService);
        when(centralConfigService.getMutableDescriptor()).thenReturn(mutableCentralConfigDescriptor);
        when(mutableCentralConfigDescriptor.getCurrentReverseProxy()).thenReturn(null);
        when(artifactoryHome.getArtifactoryProperties()).thenReturn(artifactorySystemProperties);

        when(downloadRedirectHelper.isRedirectEnableForThisRepoType(anyString())).thenReturn(true);
        when(binaryService.isCdnAllowed()).thenReturn(true);
        AolService aolService = mock(AolService.class);
        when(aolService.isAol()).thenReturn(true);
        when(artifactoryContext.beanForType(AolService.class)).thenReturn(aolService);
    }

    @AfterMethod
    private void cleanup() {
        ArtifactoryContextThreadBinder.unbind();
        ArtifactoryHome.unbind();
    }

    @Test
    public void testBuildLocalDescriptorWithCdnRedirectEnabled() {
        typeSpecificConfigModel = new NpmTypeSpecificConfigModel();
        final LocalRepositoryConfigModel model = prepareLocalModel();
        final LocalRepoDescriptor localRepoDescriptor = repoConfigDescriptorBuilder.buildLocalDescriptor(model);
        assertTrue(localRepoDescriptor.getCdnRedirectRepoConfig().isEnabled());
    }

    @Test
    public void testBuildFederatedRepoDescriptorWithProxy() {
        typeSpecificConfigModel = new NpmTypeSpecificConfigModel();
        final FederatedRepositoryConfigModel model = prepareFederatedRepoModel();
        ProxyDescriptor proxyDescriptor = new ProxyDescriptor();
        proxyDescriptor.setKey("proxy");
        CentralConfigDescriptor descriptor = mock(CentralConfigDescriptor.class);
        when(descriptor.getProxy("proxy")).thenReturn(proxyDescriptor);
        when(centralConfigService.getDescriptor()).thenReturn(descriptor);
        final FederatedRepoDescriptor federatedRepoDescriptor =
                repoConfigDescriptorBuilder.buildFederatedDescriptor(model);
        assertNotNull(federatedRepoDescriptor.getProxy());
        assertEquals(federatedRepoDescriptor.getProxy().getKey(), "proxy");
    }

    @Test
    public void testBuildRemoteDescriptor() {
        typeSpecificConfigModel = new NpmTypeSpecificConfigModel();
        final RemoteRepositoryConfigModel model = prepareRemoteModel();
        final RemoteRepoDescriptor remoteRepoDescriptor = repoConfigDescriptorBuilder.buildRemoteDescriptor(model);
        assertTrue(remoteRepoDescriptor.getCdnRedirectRepoConfig().isEnabled());
        long metadataRetrievalTimeoutSecs = model.getAdvanced().getCache().getMetadataRetrievalTimeoutSecs();
        assertEquals(remoteRepoDescriptor.getMetadataRetrievalTimeoutSecs(), metadataRetrievalTimeoutSecs);
    }

    @Test
    public void testBuildDockerRemoteDescriptorWithNewCacheRetrievalPeriod() {
        typeSpecificConfigModel = new DockerTypeSpecificConfigModel();
        final RemoteRepositoryConfigModel model = prepareRemoteModel();
        final RemoteRepoDescriptor remoteRepoDescriptor = repoConfigDescriptorBuilder.buildRemoteDescriptor(model);
        assertEquals(remoteRepoDescriptor.getRetrievalCachePeriodSecs(), DEFAULT_DOCKER_REMOTE_RETRIEVAL_CACHE_PERIOD);
    }

    @Test
    public void testPopulateVirtualBasicDescriptorValues() {
        RepoConfigDescriptorBuilder descriptorBuilder = new RepoConfigDescriptorBuilder();
        VirtualBasicRepositoryConfigModel model = mock(VirtualBasicRepositoryConfigModel.class);
        String repoName = "repo";
        when(model.getDefaultDeploymentRepo()).thenReturn(repoName);
        VirtualRepoDescriptor descriptor = new VirtualRepoDescriptor();
        MutableCentralConfigDescriptor configDescriptor = mock(MutableCentralConfigDescriptor.class);
        Map<String, LocalRepoDescriptor> reposMap = Maps.newLinkedHashMap();
        FederatedRepoDescriptor federatedRepoDescriptor = new FederatedRepoDescriptor();
        reposMap.put(repoName, federatedRepoDescriptor);
        when(configDescriptor.getLocalAndFederatedRepositoriesMap()).thenReturn(reposMap);
        descriptorBuilder.populateVirtualBasicDescriptorValues(model, descriptor, configDescriptor);
        assertEquals(descriptor.getDefaultDeploymentRepo(), federatedRepoDescriptor);
    }

    private void fillBasic(BasicRepositoryConfigModel basic) {
        basic.setPublicDescription("");
        basic.setInternalDescription("");
        basic.setIncludesPattern("");
        basic.setExcludesPattern("");
    }

    private LocalRepositoryConfigModel prepareLocalModel() {

        final LocalRepositoryConfigModel model = new LocalRepositoryConfigModel();
        CdnRedirectRepoConfigModel cdnRedirectRepoConfig = new CdnRedirectRepoConfigModel();
        cdnRedirectRepoConfig.setEnabled(true);
        final LocalAdvancedRepositoryConfigModel advanced = new LocalAdvancedRepositoryConfigModel();
        final GeneralRepositoryConfigModel general = new GeneralRepositoryConfigModel();
        final LocalBasicRepositoryConfigModel basic = new LocalBasicRepositoryConfigModel();
        fillBasic(basic);
        general.setRepoKey("repo");

        advanced.setCdnRedirectRepoConfig(cdnRedirectRepoConfig);

        model.setTypeSpecific(typeSpecificConfigModel);
        model.setAdvanced(advanced);
        model.setGeneral(general);
        model.setBasic(basic);

        return model;
    }

    private FederatedRepositoryConfigModel prepareFederatedRepoModel() {

        final FederatedRepositoryConfigModel model = new FederatedRepositoryConfigModel();

        final FederatedRepoAdvancedRepositoryConfigModel advanced = new FederatedRepoAdvancedRepositoryConfigModel();
        advanced.setProxy("proxy");
        final GeneralRepositoryConfigModel general = new GeneralRepositoryConfigModel();
        final LocalBasicRepositoryConfigModel basic = new LocalBasicRepositoryConfigModel();
        fillBasic(basic);
        general.setRepoKey("repo");
        model.setFederation(new FederationConfigModel());
        model.setTypeSpecific(typeSpecificConfigModel);
        model.setAdvanced(advanced);
        model.setGeneral(general);
        model.setBasic(basic);

        return model;
    }

    private RemoteRepositoryConfigModel prepareRemoteModel() {
        final RemoteRepositoryConfigModel model = new RemoteRepositoryConfigModel();
        CdnRedirectRepoConfigModel cdnRedirectRepoConfig = new CdnRedirectRepoConfigModel();
        cdnRedirectRepoConfig.setEnabled(true);
        final RemoteAdvancedRepositoryConfigModel advanced = new RemoteAdvancedRepositoryConfigModel();
        RemoteCacheRepositoryConfigModel cache = new RemoteCacheRepositoryConfigModel();
        advanced.setCache(cache);
        final GeneralRepositoryConfigModel general = new GeneralRepositoryConfigModel();
        final RemoteBasicRepositoryConfigModel basic = new RemoteBasicRepositoryConfigModel();
        fillBasic(basic);

        general.setRepoKey("repo");

        advanced.setCdnRedirectRepoConfig(cdnRedirectRepoConfig);

        model.setTypeSpecific(typeSpecificConfigModel);
        model.setAdvanced(advanced);
        model.setGeneral(general);
        model.setBasic(basic);
        when(artifactorySystemProperties.getProperty(any(ConstantValues.class)))
                .thenAnswer(invocationOnMock -> {
                    ConstantValues constantValue = invocationOnMock.getArgument(0);
                    if (constantValue == ConstantValues.debianDefaultArchitectures) {
                        return "deb";
                    }
                    return constantValue.getDefValue().trim();
                });

        return model;
    }
}